@@ -65,8 +65,9 @@ cpt_list.list <- function(x, g = NULL) {
   g <- graph_from_cpt_list(y)
   if (!igraph::is_dag(g)) stop("The cpts does not induce an acyclic graph.")
 
-  structure(y,
-    names     = names(x),
+  structure(
+    y,
+    nodes     = names(x),
     dim_names = dim_names,
     parents   = parents_cpt_list(y), # NOTE: Needed?
     graph     = g,
@@ -86,8 +87,12 @@ cpt_list.data.frame <- function(x, g) {
     if (!igraph::is_chordal(g)$chordal) {
       stop("undirected graphs must be be decomposable", call. = FALSE)
     }
+    message(
+      "Consider using 'pot_list' for markov random fields. ",
+      "It is faster and more idiomatic."
+    )
   }
-  
+
   parents <- if (is_dag) {
     parents_igraph(g)
   } else {
@@ -97,6 +102,7 @@ cpt_list.data.frame <- function(x, g) {
   dns <- list()
 
   y <- lapply(seq_along(parents), function(i) {
+    # TODO: Stop making the dense cpts first. Just construct the sparta object directly
     child <- names(parents)[i]
     pars  <- parents[[i]]
     spar  <- sparta::as_sparta(x[, c(child, pars), drop = FALSE])
@@ -110,8 +116,8 @@ cpt_list.data.frame <- function(x, g) {
   dns <- dns[unique(names(dns))]
   
   structure(
-    y,
-    names     = names(parents),
+    structure(y, names = names(parents)),
+    nodes     = names(parents),
     dim_names = dns,
     parents   = parents,
     graph     = g,
@@ -119,11 +125,69 @@ cpt_list.data.frame <- function(x, g) {
   )
 }
 
+
+#' A check and extraction of clique potentials from a Markov random field
+#' to be used in the junction tree algorithm
+#'
+#' @param x Character \code{data.frame}
+#' @param g A decomposable Markov random field as an igraph object.
+#' @examples
+#'
+#' # Typically one would use the ess package:
+#' # library(ess)
+#' # g  <- ess::fit_graph(derma)
+#' # pl <- pot_list(derma, ess::as_igraph(g))
+#' # pl
+#'
+#' # Another example
+#' g <- igraph::sample_gnm(ncol(asia), 12)
+#' while(!igraph::is.chordal(g)$chordal) g <- igraph::sample_gnm(ncol(asia), 12, FALSE)
+#' igraph::V(g)$name <- colnames(asia)
+#' plot(g)
+#' pot_list(asia, g)
+#' 
+#' @export
+pot_list <- function(x, g) UseMethod("pot_list")
+
+
+#' @rdname pot_list
+#' @export
+pot_list.data.frame <- function(x, g) {
+
+  if (!igraph::is.igraph(g)) stop("g must be an igraph object", call. = FALSE)
+  adj_lst <- as_adj_lst(igraph::as_adjacency_matrix(g, sparse = FALSE))
+  cliques <- rip(adj_lst, check = TRUE)$C
+  names(cliques) <- paste("C", 1:length(cliques), sep = "")
+  dns <- list()
+
+  y <- lapply(seq_along(cliques), function(i) {
+    clique <- cliques[[i]]
+    spar   <- sparta::as_sparta(x[, clique, drop = FALSE])
+    spar   <- sparta::normalize(spar)
+    # This ensures, that the potentials and dim_names have the same ordering of the lvls!
+    dns   <<- push(dns, sparta::dim_names(spar))
+    spar
+  })
+
+  dns <- unlist(dns, FALSE)
+  dns <- dns[unique(names(dns))]
+  
+  structure(
+    structure(y, names = names(cliques)),
+    nodes     = colnames(x),
+    cliques   = cliques,
+    dim_names = dns,
+    graph     = g,
+    class     = c("pot_list", "list")
+  )
+}
+
 #' Compile information
 #'
 #' Compiled objects are used as building blocks for junction tree inference
 #'
-#' @param x An object returned from \code{cpt_list}
+#' @param x An object returned from \code{cpt_list} (baeysian network) or
+#' \code{pot_list} (decomposable markov random field)
 #' @param evidence A named vector. The names are the variabes and the elements
 #' are the evidence.
 #' @param root_node A node for which we require it to live in the root
@@ -131,10 +195,21 @@ cpt_list.data.frame <- function(x, g) {
 #' @param joint_vars A vector of variables for which we require them
 #' to be in the same clique. Edges between all these variables are added
 #' to the moralized graph.
-#' @param tri The optimization strategy used for triangulation. One of
-#' 'min_nei', 'min_fill', 'min_rfill', 'min_sp', 'evidence', 'minimal', 'alpha'.
+#' @param tri The optimization strategy used for triangulation if x originates
+#' from a Baeysian network. One of
+#'  * 'min_fill'
+#'  * 'min_rfill'
+#'  * 'min_efill'
+#'  * 'min_sfill'
+#'  * 'min_sp'
+#'  * 'min_esp'
+#'  * 'min_nei'
+#'  * 'minimal'
+#'  * 'alpha'
+#' @md
 #' @param pmf_evidence A named vector of frequencies. The names should
-#' correspond to the evidence that is expected to see over time.
+#' correspond to the evidence that is expected to see over time. Relevant
+#' in connection to \code{min_efill} and \code{min_esp} triangulations.
 #' @param alpha Character vector. A permutation of the nodes
 #' in the graph. It specifies a user-supplied eliminination ordering for
 #' triangulation of the moral graph.
@@ -160,6 +235,12 @@ cpt_list.data.frame <- function(x, g) {
 #' after compilation. Before refers to entering
 #' evidence in the 'compile' function and after
 #' refers to entering evidence in the 'jt' function.
+#'
+#' Finally, one can either use a Bayesian network or a decomposable
+#' Markov random field (use the \code{ess} package to fit these). Bayesian
+#' networks must be constructed with \code{cpt_list} and decomposable MRFs
+#' should be constructed with \code{pot_list}, but can also be constructed
+#' using \code{cpt_list}.
 #' 
 #' @examples
 #' cptl <- cpt_list(asia2)
@@ -191,7 +272,7 @@ compile.cpt_list <- function(x,
                              alpha          = NULL                             
                              ) {
 
-  .defense_compile(tri, pmf_evidence, alpha, names(x))
+  check_params_compile(tri, pmf_evidence, alpha, names(x), root_node)
   
   g       <- attr(x, "graph")
   parents <- attr(x, "parents")
@@ -203,68 +284,142 @@ compile.cpt_list <- function(x,
   M  <- igraph::as_adjacency_matrix(gm, sparse = FALSE)
   
   tri_obj <- switch(tri,
-    "min_fill"  = new_min_fill_triang(M),
-    "min_rfill" = new_min_rfill_triang(M),
-    "min_sp"    = new_min_sp_triang(M, .map_int(dim_names(x), length)),
-    "min_nei"   = new_min_nei_triang(M),
-    "minimal"   = new_minimal_triang(M),
-    "evidence"  = new_evidence_triang(M, .map_int(dim_names(x), length), pmf_evidence),
-    # "evidence2" = new_evidence2_triang(M, .map_int(dim_names(x), length), pmf_evidence),
-    "alpha"     = new_alpha_triang(M, alpha)
+    "min_fill"   = new_min_fill_triang(M),
+    "min_rfill"  = new_min_rfill_triang(M),
+    "min_efill"  = new_min_efill_triang(M, .map_int(dim_names(x), length), pmf_evidence),
+    "min_sfill"  = new_min_sfill_triang(M, .map_int(dim_names(x), length)),
+    "min_rsfill" = new_min_rsfill_triang(M, .map_int(dim_names(x), length)),
+    "min_sp"     = new_min_sp_triang(M, .map_int(dim_names(x), length)),
+    "min_esp"    = new_min_esp_triang(M, .map_int(dim_names(x), length), pmf_evidence),
+    "min_nei"    = new_min_nei_triang(M),
+    "minimal"    = new_minimal_triang(M),
+    "alpha"      = new_alpha_triang(M, alpha)
   )
   
   gmt     <- .triang(tri_obj)
   adj_lst <- as_adj_lst(gmt)
-  cliques <- construct_cliques(adj_lst, root_node) # just use cliques_int ?
+  cliques <- construct_cliques(adj_lst)
 
   # cliques_int is needed to construct the junction tree in new_jt -> new_schedule
   dimnames(gmt) <- lapply(dimnames(gmt), function(x) 1:nrow(gmt))
   adj_lst_int   <- as_adj_lst(gmt)
-  root_node_int <- ifelse(root_node != "", as.character(match(root_node, names(adj_lst))), "")
-  cliques_int   <- lapply(rip(adj_lst_int, root_node_int)$C, as.integer)
+  # root_node_int <- ifelse(root_node != "", as.character(match(root_node, names(adj_lst))), "")
+  cliques_int   <- lapply(rip(adj_lst_int)$C, as.integer)
 
+  # TODO: REMOVE THE EVIDENCE VARIABLES IN g, IN ADVANCE
+  # AND EXPLOIT THE NEW DAG STRUCTURE! Removing "a" in p(a|b) should result in a
+  # new cpt with no parents but a child var of "a". Then things are consistent again.
+  
+  inc <- new.env()
+  inc$inc <- FALSE
   if (!is.null(evidence)) {
     if (!valid_evidence(attr(x, "dim_names"), evidence)) {
       stop("Evidence is not on correct form", call. = FALSE)
     }
-
     # x looses its attributes in set_evidence
     att_ <- attributes(x)
-    x    <- set_evidence_(x, cliques, evidence)
+    x    <- set_evidence_(x, evidence, inc)
     attributes(x) <- att_
   }
- 
-  charge  <- new_charge(x, cliques, parents)
-  out     <- structure(
-    list(charge = charge, cliques = cliques),
-    root_node   = root_node,
-    dim_names   = attr(x, "dim_names"),
-    evidence    = evidence,
-    graph       = g,
-    cliques_int = cliques_int,
-    class       = c("charge", "list")
-  )
 
-  out
+  charge  <- new_charge_cpt(x, cliques, parents)
+
+  structure(
+    list(charge   = charge, cliques = cliques),
+    root_node     = root_node,
+    joint_vars    = joint_vars,
+    dim_names     = attr(x, "dim_names"),
+    evidence      = evidence,
+    graph         = g,
+    cliques_int   = cliques_int,
+    inconsistencies = inc$inc,
+    class         = c("charge", "list")
+  )
 }
 
-#' Variable getters
+
+#' @rdname compile
+#' @export
+compile.pot_list <- function(x,
+                             evidence       = NULL,
+                             root_node      = "",
+                             joint_vars     = NULL,
+                             tri            = "min_fill",
+                             pmf_evidence   = NULL,
+                             alpha          = NULL                             
+                             ) {
+
+  check_params_compile(tri, pmf_evidence, alpha, names(x), root_node)
+  
+  g       <- attr(x, "graph")
+  gmat    <- igraph::as_adjacency_matrix(g, sparse = FALSE)
+  adj_lst <- as_adj_lst(gmat)
+  cliques <- attr(x, "cliques")
+
+  # cliques_int is needed to construct the junction tree in new_jt -> new_schedule
+  dimnames(gmat) <- lapply(dimnames(gmat), function(x) 1:nrow(gmat))
+  adj_lst_int   <- as_adj_lst(gmat)
+  # root_node_int <- ifelse(root_node != "", as.character(match(root_node, names(adj_lst))), "")
+  cliques_int   <- lapply(rip(adj_lst_int)$C, as.integer)
+
+  inc <- new.env()
+  inc$inc <- FALSE
+  if (!is.null(evidence)) {
+    if (!valid_evidence(attr(x, "dim_names"), evidence)) {
+      stop("Evidence is not on correct form", call. = FALSE)
+    }
+    # x looses its attributes in set_evidence
+    att_ <- attributes(x)
+    x    <- set_evidence_(x, evidence, inc)
+    attributes(x) <- att_
+  }
+
+  charge  <- new_charge_pot(x)
+  structure(
+    list(charge   = charge, cliques = cliques),
+    root_node     = root_node,
+    joint_vars    = joint_vars,
+    dim_names     = attr(x, "dim_names"),
+    evidence      = evidence,
+    graph         = g,
+    cliques_int   = cliques_int,
+    inconsistencies = inc$inc,
+    class         = c("charge", "list")
+  )
+}
+
+
+
+#' Various getters
 #'
-#' Getter methods for information about variables and their statespaces
+#' Getter methods for \code{cpt_list}, \code{pot_list}, \code{charge}
+#' and \code{jt} objects
 #' 
-#' @param x \code{cpt_list} or a compiled object
+#' @param x \code{cpt_list}, \code{pot_list}, \code{charge} or \code{jt}
 
 #' @rdname getters
 #' @export
 dim_names <- function(x) UseMethod("dim_names")
 
+#' @rdname getters
+#' @export
+has_inconsistencies <- function(x) UseMethod("has_inconsistencies")
+
 #' @rdname getters
 #' @export
 dim_names.cpt_list <- function(x) attr(x, "dim_names")
 
 #' @rdname getters
 #' @export
-names.cpt_list <- function(x) attr(x, "names")
+names.cpt_list <- function(x) attr(x, "nodes")
+
+#' @rdname getters
+#' @export
+dim_names.pot_list <- function(x) attr(x, "dim_names")
+
+#' @rdname getters
+#' @export
+names.pot_list <- function(x) attr(x, "nodes")
 
 #' @rdname getters
 #' @export
@@ -274,6 +429,10 @@ dim_names.charge <- function(x) attr(x, "dim_names")
 #' @export
 names.charge <- function(x) names(attr(x, "dim_names"))
 
+#' @rdname getters
+#' @export
+has_inconsistencies.charge <- function(x) attr(x, "inconsistencies")
+
 #' @rdname getters
 #' @export
 dim_names.jt <- function(x) attr(x, "dim_names")
@@ -282,6 +441,9 @@ dim_names.jt <- function(x) attr(x, "dim_names")
 #' @export
 names.jt <- function(x) names(attr(x, "dim_names"))
 
+#' @rdname getters
+#' @export
+has_inconsistencies.jt <- function(x) attr(x, "inconsistencies")
 
 #' Get graph
 #'
@@ -302,6 +464,9 @@ get_graph.charge <- function(x) attr(x, "graph")
 #' @export
 get_graph.cpt_list <- function(x) attr(x, "graph")
 
+#' @rdname get-graph
+#' @export
+get_graph.pot_list <- function(x) attr(x, "graph")
 
 #' A print method for cpt lists
 #'
@@ -327,9 +492,31 @@ print.cpt_list <- function(x, ...) {
   cat(paste0("\n  ", cls),
     "\n -------------------------\n"
   )
-  
 }
 
+#' A print method for pot lists
+#'
+#' @param x A \code{pot_list} object
+#' @param ... For S3 compatability. Not used.
+#' @seealso \code{\link{compile}}
+#' @export
+print.pot_list <- function(x, ...) {
+  cls <- paste0("<", paste0(class(x), collapse = ", "), ">")
+  nn  <- length(names(x))
+  cat(" List of potentials",
+    "\n -------------------------\n")
+
+  for (k in seq_along(x)) {
+    pot_vars <- names(x[[k]])
+    cat("  pot(", paste(pot_vars, collapse = ", "), ")\n")      
+  }
+  
+  cat(paste0("\n  ", cls),
+    "\n -------------------------\n"
+  )
+}
+
+
 #' A print method for compiled objects
 #'
 #' @param x A compiled object
@@ -337,7 +524,6 @@ print.cpt_list <- function(x, ...) {
 #' @seealso \code{\link{jt}}
 #' @export
 print.charge <- function(x, ...) {
-  # TODO: PRINT EVIDENCE
   cls <- paste0("<", paste0(class(x), collapse = ", "), ">")
   nn  <- length(names(x))
   clique_sizes <- .map_int(x$cliques, length)
@@ -354,8 +540,9 @@ print.charge <- function(x, ...) {
   )
 
   e <- attr(x, "evidence")
+  inc <- attr(x, "inconsistencies")
   if (!is.null(e)) {
-    cat("\n  Evidence:")
+    if (inc) cat("\n  Evidence: (inconsistencies)") else cat("\n  Evidence:")
     for (i in seq_along(e)) {
       cat(
         "\n   -", paste0(names(e[i]), ":"), unname(e[i])
